"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GolangGenerator = void 0;
var templates = __importStar(require("./templates"));
var liquidjs_1 = require("liquidjs");
var graphql_1 = require("graphql");
/**
 * GolangGenerator is a class that you instanciate with a GraphQL schema
 * and an optional configuration. You can then call its method `generate`
 * to generate Golang code.
 */
var GolangGenerator = /** @class */ (function () {
    function GolangGenerator(schema, config) {
        var _this = this;
        /**
         * Map GraphQL type to their Golang format.
         */
        this.types = {
            Int: 'Int',
            Float: 'Float',
            Boolean: 'Boolean',
            String: 'String',
            ID: 'ID',
        };
        /**
         * Map GraphQL scalar types to Golang types.
         */
        this.scalars = {
            Int: 'int32',
            Float: 'float64',
            Boolean: 'bool',
            String: 'string',
            ID: 'string',
        };
        this._enums = [];
        this._inputs = [];
        this._objects = [];
        this.config = config !== null && config !== void 0 ? config : {};
        this.schema = schema;
        this._templateEngine = new liquidjs_1.Liquid({
            strictFilters: true,
            strictVariables: true,
        });
        this._templateOperation = this._templateEngine.parse(templates.GOLANG_OPERATION);
        (0, graphql_1.visit)((0, graphql_1.parse)((0, graphql_1.printSchema)(this.schema)), {
            ScalarTypeDefinition: function (node) {
                _this.types[node.name.value] = _this._formatName(node.name.value);
                _this.scalars[node.name.value] = 'string';
                return false;
            },
            EnumTypeDefinition: function (node) {
                _this.types[node.name.value] = _this._formatName(node.name.value);
                _this._enums.push(node);
                return false;
            },
            InputObjectTypeDefinition: function (node) {
                _this.types[node.name.value] = _this._formatName(node.name.value);
                _this._inputs.push(node);
                return false;
            },
            ObjectTypeDefinition: function (node) {
                _this.types[node.name.value] = _this._formatName(node.name.value);
                _this._objects.push(node);
                return false;
            },
        });
    }
    GolangGenerator.prototype.generate = function (documents) {
        var _this = this;
        var l = __spreadArray(__spreadArray(__spreadArray([], this._generatePackage(), true), this._generateImports(), true), [
            templates.GOLANG_BASE,
        ], false);
        documents === null || documents === void 0 ? void 0 : documents.forEach(function (document) {
            return l.push.apply(l, _this._generateOperations(document));
        });
        l.push.apply(l, this._generateSchema());
        return l.join('\n');
    };
    /**
     * Format a name for Golang.
     * @param name - Name to format.
     */
    GolangGenerator.prototype._formatName = function (name) {
        if (name.match(/^id$/i) || name.match(/^uuid$/i)) {
            return name.toUpperCase();
        }
        return name
            .replace(/(^_|_$)/, '')
            .split('_')
            .map(function (word) { return word[0].toUpperCase() + word.substr(1); })
            .join('');
    };
    /**
     * Generate Golang package header.
     */
    GolangGenerator.prototype._generatePackage = function () {
        var _a;
        return [
            "package ".concat((_a = this.config.packageName) !== null && _a !== void 0 ? _a : 'graphql'),
            '',
            '// Code generated by graphql-codegen-golang ; DO NOT EDIT.',
            '',
        ];
    };
    /**
     * Generate Golang imports for operations.
     */
    GolangGenerator.prototype._generateImports = function () {
        return [
            'import (',
            '  "bytes"',
            '  "encoding/json"',
            '  "fmt"',
            '  "io/ioutil"',
            '  "net/http"',
            '  "strings"',
            ')',
            '',
        ];
    };
    /**
     * Generate a named comment section.
     * @param name - Name of the section.
     */
    GolangGenerator.prototype._generateSection = function (name) {
        return ['', '//', "// ".concat(name), '//', ''];
    };
    /**
     * Generate the Golang types from schema.
     */
    GolangGenerator.prototype._generateSchema = function () {
        return __spreadArray(__spreadArray(__spreadArray(__spreadArray([], this._generateScalars(), true), this._generateEnums(), true), this._generateInputs(), true), this._generateObjects(), true);
    };
    GolangGenerator.prototype._generateScalars = function () {
        var _this = this;
        var l = __spreadArray([], this._generateSection('Scalars'), true);
        Object.entries(this.scalars).map(function (_a) {
            var gqlType = _a[0], goType = _a[1];
            l.push("type ".concat(_this.types[gqlType], " ").concat(goType));
        });
        return l;
    };
    GolangGenerator.prototype._generateEnums = function () {
        var _this = this;
        var l = __spreadArray([], this._generateSection('Enums'), true);
        this._enums.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type ".concat(_this.types[node.name.value], " string"), 'const (');
            (_a = node.values) === null || _a === void 0 ? void 0 : _a.forEach(function (value) {
                var name = value.name.value;
                l.push("  ".concat(goType).concat(_this._formatName(name), " ").concat(goType, " = \"").concat(name, "\""));
            });
            l.push(')');
        });
        return l;
    };
    GolangGenerator.prototype._generateInputs = function () {
        var _this = this;
        var l = __spreadArray([], this._generateSection('Inputs'), true);
        this._inputs.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type ".concat(goType, " struct {"));
            (_a = node.fields) === null || _a === void 0 ? void 0 : _a.forEach(function (field) {
                l.push(_this._generateField(field.name.value, field.type));
            });
            l.push('}');
        });
        return l;
    };
    GolangGenerator.prototype._generateObjects = function () {
        var _this = this;
        var l = __spreadArray([], this._generateSection('Objects'), true);
        this._objects.forEach(function (node) {
            var _a;
            var goType = _this.types[node.name.value];
            l.push('', "type ".concat(goType, " struct {"));
            (_a = node.fields) === null || _a === void 0 ? void 0 : _a.forEach(function (field) {
                l.push(_this._generateField(field.name.value, field.type));
            });
            l.push('}');
        });
        return l;
    };
    /**
     * Generate a Golang struct's field.
     * @param name - Name of the GraphQL field.
     * @param type - Type of the GraphQL field.
     */
    GolangGenerator.prototype._generateField = function (name, type) {
        return "  ".concat(this._formatName(name), " ").concat(this._generateFieldType(type, name));
    };
    /**
     *
     * @param type - TypeNode contains a GraphQL type.
     * @param fieldName - GraphQL field's name for the Golang's json annotation.
     * @param prefix - String accumulator to add to result when recursion is done.
     * @param nonNull - Boolean accumulator to determine if type must be a pointer.
     */
    GolangGenerator.prototype._generateFieldType = function (type, fieldName, prefix, nonNull) {
        if (prefix === void 0) { prefix = ''; }
        if (nonNull === void 0) { nonNull = false; }
        if (type.kind === 'NamedType') {
            return [
                "".concat(nonNull ? '' : '*').concat(prefix),
                "".concat(this.types[type.name.value], " "),
                "`json:\"".concat(fieldName).concat(nonNull ? '' : ',omitempty', "\"`"),
            ].join('');
        }
        if (type.kind === 'NonNullType') {
            return this._generateFieldType(type.type, fieldName, prefix, prefix !== '' ? false : true);
        }
        if (type.kind === 'ListType') {
            return this._generateFieldType(type.type, fieldName, prefix + '[]', nonNull);
        }
        throw new Error("field type \"".concat(type, "\" not supported!"));
    };
    /**
     * Generate operations code from a GraphQL document.
     * @param document - Document to get the operations from.
     */
    GolangGenerator.prototype._generateOperations = function (document) {
        var _this = this;
        var l = [];
        var typeInfo = new graphql_1.TypeInfo(this.schema);
        (0, graphql_1.visit)(document, (0, graphql_1.visitWithTypeInfo)(typeInfo, {
            OperationDefinition: {
                enter: function (operation) {
                    // Anonymous operation are not supported: skip them.
                    if (!operation.name)
                        return false;
                    // Subscription operations are not supported yes.
                    if (operation.operation === 'subscription')
                        return false;
                    var name = _this._formatName(operation.name.value);
                    l.push.apply(l, _this._generateSection("".concat((0, graphql_1.print)(operation).split('{', 1)[0].trim())));
                    // Generate operation variables type if any
                    l.push.apply(l, _this._generateOperationVariables(name, operation.variableDefinitions));
                    // Finally, operation's response type are generated when visiting
                    // inner AST nodes
                    l.push("type ".concat(name, "Response struct {"));
                },
                leave: function (operation) {
                    if (!operation.name)
                        return;
                    // Inner nodes have been visited. Time to close response type.
                    l.push('}', '');
                    var name = _this._formatName(operation.name.value);
                    var hasVariables = _this._generateOperationVariables(name, operation.variableDefinitions).length > 0;
                    // Generate everything except variables and response type
                    l.push.apply(l, _this._generateOperationCode(name, (0, graphql_1.print)(operation), hasVariables));
                },
            },
            Field: {
                enter: function (field) {
                    var _a;
                    var name = _this._formatName(field.name.value);
                    var w = ["  ".concat(name, " ")];
                    if (field.selectionSet) {
                        var outputType = (_a = typeInfo.getType()) === null || _a === void 0 ? void 0 : _a.toString();
                        if (outputType === null || outputType === void 0 ? void 0 : outputType.startsWith('[')) {
                            if (!outputType.endsWith('!')) {
                                w.push('*');
                            }
                            w.push('[]');
                        }
                        w.push('struct {');
                    }
                    else {
                        w.push("string `json:\"".concat(field.name.value, "\"`"));
                    }
                    l.push(w.join(''));
                },
                leave: function (field) {
                    if (field.selectionSet) {
                        l.push("} `json:\"".concat(field.name.value, "\"`"));
                    }
                },
            },
        }));
        return l;
    };
    /**
     * Generate Golang `type ${name}Variables struct {...}`.
     * @param name - Name of the operation. Must be formatted for Golang.
     * @param variableDefinitions - GraphQL variables to generate Golang fields from.
     */
    GolangGenerator.prototype._generateOperationVariables = function (name, variableDefinitions) {
        var _this = this;
        if (!variableDefinitions || variableDefinitions.length === 0) {
            return [];
        }
        var l = ["type ".concat(name, "Variables struct {")];
        variableDefinitions.forEach(function (variable) {
            l.push(_this._generateField(variable.variable.name.value, variable.type));
        });
        l.push('}', '');
        return l;
    };
    /**
     * Generates all Golang code for an operation except `type ${name}Variables`
     * and `type ${name}Response`.
     * @param name - Name of the operation. Must be formatted for Golang.
     * @param operation - String version of the operation.
     * @param hasVariables - Whether the operation uses a `${name}Variables` or not.
     */
    GolangGenerator.prototype._generateOperationCode = function (name, operation, hasVariables) {
        var l = [];
        l.push(this._templateEngine.renderSync(this._templateOperation, {
            name: name,
            operation: operation,
            hasVariables: hasVariables,
        }));
        return l;
    };
    return GolangGenerator;
}());
exports.GolangGenerator = GolangGenerator;
